import logging
from pathlib import Path

from databricks.labs.remorph.config import get_dialect
from databricks.labs.remorph.helpers.file_utils import (
    get_sql_file,
    is_sql_file,
    read_file,
)
from databricks.labs.remorph.intermediate.dag import DAG
from databricks.labs.remorph.intermediate.engine_adapter import EngineAdapter

logger = logging.getLogger(__name__)


class RootTableIdentifier:
    def __init__(self, source: str, input_path: str | Path):
        self.source = source
        self.input_path = input_path
        self.engine_adapter = EngineAdapter(get_dialect(source))

    def generate_lineage(self, engine="sqlglot") -> DAG:
        dag = DAG()

        # when input is sql file then parse the file
        if is_sql_file(self.input_path):
            filename = self.input_path
            logger.debug(f"Generating Lineage file: {filename}")
            sql_content = read_file(filename)
            self.engine_adapter.parse_sql_content(dag, sql_content, filename, engine)
            return dag  # return after processing the file

        # when the input is a directory
        for filename in get_sql_file(self.input_path):
            logger.debug(f"Generating Lineage file: {filename}")
            sql_content = read_file(filename)
            self.engine_adapter.parse_sql_content(dag, sql_content, filename, engine)

        return dag



import logging
import os

from databricks.labs.lsql.backends import (
    DatabricksConnectBackend,
    RuntimeBackend,
    SqlBackend,
    StatementExecutionBackend,
)
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


def get_sql_backend(ws: WorkspaceClient, warehouse_id: str | None = None) -> SqlBackend:
    warehouse_id = warehouse_id or ws.config.warehouse_id
    if warehouse_id:
        sql_backend: SqlBackend = StatementExecutionBackend(ws, warehouse_id)
    else:
        sql_backend = RuntimeBackend() if "DATABRICKS_RUNTIME_VERSION" in os.environ else DatabricksConnectBackend(ws)
    return sql_backend



from databricks.sdk.core import with_user_agent_extra, with_product
from databricks.labs.blueprint.logger import install_logger
from databricks.labs.remorph.__about__ import __version__

install_logger()

# Add remorph/<version> for projects depending on remorph as a library
with_user_agent_extra("remorph", __version__)

# Add remorph/<version> for re-packaging of remorph, where product name is omitted
with_product("remorph", __version__)



import logging
from sqlglot import expressions as exp

from databricks.labs.remorph.snow import databricks as extended_databricks

logger = logging.getLogger(__name__)


class DatabricksExperimental(extended_databricks.Databricks):
    databricks = extended_databricks.Databricks()

    class Generator(extended_databricks.Databricks.Generator):
        def _parse_json(self, expression: exp.ParseJSON) -> str:
            return self.func("PARSE_JSON", expression.this, expression.expression)

        TRANSFORMS = {
            **extended_databricks.Databricks.Generator.TRANSFORMS,
            exp.ParseJSON: _parse_json,
        }

        TYPE_MAPPING = {
            **extended_databricks.Databricks.Generator.TYPE_MAPPING,
            exp.DataType.Type.VARIANT: "VARIANT",
        }



import logging

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound

from databricks.labs.remorph.config import RemorphConfigs
from databricks.labs.remorph.deployment.recon import ReconDeployment

logger = logging.getLogger("databricks.labs.remorph.install")


class WorkspaceInstallation:
    def __init__(
        self,
        ws: WorkspaceClient,
        prompts: Prompts,
        installation: Installation,
        recon_deployment: ReconDeployment,
    ):
        self._ws = ws
        self._prompts = prompts
        self._installation = installation
        self._recon_deployment = recon_deployment

    def install(self, config: RemorphConfigs):
        if config.reconcile:
            self._recon_deployment.install(config.reconcile)

    def uninstall(self, config: RemorphConfigs):
        # This will remove all the Remorph modules
        if not self._prompts.confirm(
            "Do you want to uninstall Remorph from the workspace too, this would "
            "remove Remorph project folder, jobs, metadata and dashboards"
        ):
            return
        logger.info(f"Uninstalling Remorph from {self._ws.config.host}.")
        try:
            self._installation.files()
        except NotFound:
            logger.error(f"Check if {self._installation.install_folder()} is present. Aborting uninstallation.")
            return

        if config.morph:
            logging.info(
                f"Won't remove transpile validation schema `{config.morph.schema_name}` "
                f"from catalog `{config.morph.catalog_name}`. Please remove it manually."
            )

        if config.reconcile:
            self._recon_deployment.uninstall(config.reconcile)

        self._installation.remove()
        logger.info("Uninstallation completed successfully.")



import logging
from importlib.resources import files

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InvalidParameterValue, NotFound

import databricks.labs.remorph.resources
from databricks.labs.remorph.config import ReconcileConfig
from databricks.labs.remorph.deployment.dashboard import DashboardDeployment
from databricks.labs.remorph.deployment.job import JobDeployment
from databricks.labs.remorph.deployment.table import TableDeployment

logger = logging.getLogger(__name__)

_RECON_PREFIX = "Reconciliation"
RECON_JOB_NAME = f"{_RECON_PREFIX} Runner"
RECON_METRICS_DASHBOARD_NAME = f"{_RECON_PREFIX} Metrics"


class ReconDeployment:
    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        table_deployer: TableDeployment,
        job_deployer: JobDeployment,
        dashboard_deployer: DashboardDeployment,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._table_deployer = table_deployer
        self._job_deployer = job_deployer
        self._dashboard_deployer = dashboard_deployer

    def install(self, recon_config: ReconcileConfig | None):
        if not recon_config:
            return
        logger.info("Installing reconcile components.")
        self._deploy_tables(recon_config)
        self._deploy_dashboards(recon_config)
        self._deploy_jobs(recon_config)
        self._install_state.save()
        logger.info("Installation of reconcile components completed successfully.")

    def uninstall(self, recon_config: ReconcileConfig | None):
        if not recon_config:
            return
        logger.info("Uninstalling reconcile components.")
        self._remove_dashboards()
        self._remove_jobs()
        logging.info(
            f"Won't remove reconcile metadata schema `{recon_config.metadata_config.schema}` "
            f"from catalog `{recon_config.metadata_config.catalog}`. Please remove it and the tables inside manually."
        )
        logging.info(
            f"Won't remove configured reconcile secret scope `{recon_config.secret_scope}`. "
            f"Please remove it manually."
        )

    def _deploy_tables(self, recon_config: ReconcileConfig):
        logger.info("Deploying reconciliation metadata tables.")
        catalog = recon_config.metadata_config.catalog
        schema = recon_config.metadata_config.schema
        resources = files(databricks.labs.remorph.resources)
        query_dir = resources.joinpath("reconcile/queries/installation")

        sqls_to_deploy = [
            "main.sql",
            "metrics.sql",
            "details.sql",
            "aggregate_metrics.sql",
            "aggregate_details.sql",
            "aggregate_rules.sql",
        ]

        for sql_file in sqls_to_deploy:
            table_sql_file = query_dir.joinpath(sql_file)
            self._table_deployer.deploy_table_from_ddl_file(catalog, schema, sql_file.strip(".sql"), table_sql_file)

    def _deploy_dashboards(self, recon_config: ReconcileConfig):
        logger.info("Deploying reconciliation dashboards.")
        self._deploy_recon_metrics_dashboard(RECON_METRICS_DASHBOARD_NAME, recon_config)
        for dashboard_name, dashboard_id in self._get_deprecated_dashboards():
            try:
                logger.info(f"Removing dashboard_id={dashboard_id}, as it is no longer needed.")
                del self._install_state.dashboards[dashboard_name]
                self._ws.lakeview.trash(dashboard_id)
            except (InvalidParameterValue, NotFound):
                logger.warning(f"Dashboard `{dashboard_name}` doesn't exist anymore for some reason.")
                continue

    def _deploy_recon_metrics_dashboard(self, name: str, recon_config: ReconcileConfig):
        dashboard_params = {
            "catalog": recon_config.metadata_config.catalog,
            "schema": recon_config.metadata_config.schema,
        }

        reconcile_dashboard_path = "reconcile/dashboards/Remorph-Reconciliation.lvdash.json"
        dashboard_file = files(databricks.labs.remorph.resources).joinpath(reconcile_dashboard_path)
        logger.info(f"Creating Reconciliation Dashboard `{name}`")
        self._dashboard_deployer.deploy(name, dashboard_file, parameters=dashboard_params)

    def _get_dashboards(self) -> list[tuple[str, str]]:
        return [
            (dashboard_name, dashboard_id)
            for dashboard_name, dashboard_id in self._install_state.dashboards.items()
            if dashboard_name.startswith(_RECON_PREFIX)
        ]

    def _get_deprecated_dashboards(self) -> list[tuple[str, str]]:
        return [
            (dashboard_name, dashboard_id)
            for dashboard_name, dashboard_id in self._install_state.dashboards.items()
            if dashboard_name.startswith(_RECON_PREFIX) and dashboard_name != RECON_METRICS_DASHBOARD_NAME
        ]

    def _remove_dashboards(self):
        logger.info("Removing reconciliation dashboards.")
        for dashboard_name, dashboard_id in self._get_dashboards():
            try:
                logger.info(f"Removing dashboard {dashboard_name} with dashboard_id={dashboard_id}.")
                del self._install_state.dashboards[dashboard_name]
                self._ws.lakeview.trash(dashboard_id)
            except (InvalidParameterValue, NotFound):
                logger.warning(f"Dashboard `{dashboard_name}` doesn't exist anymore for some reason.")
                continue

    def _deploy_jobs(self, recon_config: ReconcileConfig):
        logger.info("Deploying reconciliation jobs.")
        self._job_deployer.deploy_recon_job(RECON_JOB_NAME, recon_config)
        for job_name, job_id in self._get_deprecated_jobs():
            try:
                logger.info(f"Removing job_id={job_id}, as it is no longer needed.")
                del self._install_state.jobs[job_name]
                self._ws.jobs.delete(job_id)
            except (InvalidParameterValue, NotFound):
                logger.warning(f"{job_name} doesn't exist anymore for some reason.")
                continue

    def _get_jobs(self) -> list[tuple[str, int]]:
        return [
            (job_name, int(job_id))
            for job_name, job_id in self._install_state.jobs.items()
            if job_name.startswith(_RECON_PREFIX)
        ]

    def _get_deprecated_jobs(self) -> list[tuple[str, int]]:
        return [
            (job_name, int(job_id))
            for job_name, job_id in self._install_state.jobs.items()
            if job_name.startswith(_RECON_PREFIX) and job_name != RECON_JOB_NAME
        ]

    def _remove_jobs(self):
        logger.info("Removing Reconciliation Jobs.")
        for job_name, job_id in self._get_jobs():
            try:
                logger.info(f"Removing job {job_name} with job_id={job_id}.")
                del self._install_state.jobs[job_name]
                self._ws.jobs.delete(int(job_id))
            except (InvalidParameterValue, NotFound):
                logger.warning(f"{job_name} doesn't exist anymore for some reason.")
                continue



import logging
import time

from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.catalog import Privilege
from databricks.sdk.service.sql import (
    CreateWarehouseRequestWarehouseType,
    EndpointInfoWarehouseType,
    SpotInstancePolicy,
)

from databricks.labs.remorph.helpers.metastore import CatalogOperations

logger = logging.getLogger(__name__)


class ResourceConfigurator:
    """
    Handles the setup of common Databricks resources like
    catalogs, schemas, volumes, and warehouses used across remorph modules.
    """

    def __init__(self, ws: WorkspaceClient, prompts: Prompts, catalog_ops: CatalogOperations):
        self._ws = ws
        self._user = ws.current_user.me()
        self._prompts = prompts
        self._catalog_ops = catalog_ops

    def prompt_for_catalog_setup(
        self,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_CATALOG},
        ),
        max_attempts: int = 3,
    ) -> str:
        for _ in range(1, max_attempts + 1):
            catalog_name = self._prompts.question("Enter catalog name", default="remorph")
            catalog = self._catalog_ops.get_catalog(catalog_name)
            if catalog:
                logger.info(f"Found existing catalog `{catalog_name}`")
                user_name = self._user.user_name
                assert user_name is not None
                if self._catalog_ops.has_catalog_access(catalog, user_name, required_privileges):
                    return catalog_name
                logger.info(f"User `{user_name}` doesn't have privilege to use catalog `{catalog_name}`")
                if not self._prompts.confirm("Do you want to use another catalog?"):
                    raise SystemExit("Aborting the installation.")
            else:
                if self._prompts.confirm(f"Catalog `{catalog_name}` doesn't exist. Create it?"):
                    result = self._catalog_ops.create_catalog(catalog_name)
                    assert result.name is not None
                    return result.name
                raise SystemExit("Aborting the installation.")
        raise SystemExit(f"Couldn't get answer within {max_attempts} attempts. Aborting the installation.")

    def prompt_for_schema_setup(
        self,
        catalog: str,
        default_schema_name: str,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_SCHEMA},
        ),
        max_attempts: int = 3,
    ) -> str:
        for _ in range(1, max_attempts + 1):
            schema_name = self._prompts.question("Enter schema name", default=default_schema_name)
            schema = self._catalog_ops.get_schema(catalog, schema_name)
            if schema:
                logger.info(f"Found existing schema `{schema_name}` in catalog `{catalog}`")
                user_name = self._user.user_name
                assert user_name is not None
                if self._catalog_ops.has_schema_access(schema, user_name, required_privileges):
                    return schema_name
                logger.info(
                    f"User `{user_name}` doesn't have privilege to use schema `{schema_name}` in catalog `{catalog}`"
                )
                if not self._prompts.confirm("Do you want to use another schema?"):
                    raise SystemExit("Aborting the installation.")
            else:
                if self._prompts.confirm(f"Schema `{schema_name}` doesn't exist in catalog `{catalog}`. Create it?"):
                    result = self._catalog_ops.create_schema(schema_name, catalog)
                    assert result.name is not None
                    return result.name
                raise SystemExit("Aborting the installation.")
        raise SystemExit(f"Couldn't get answer within {max_attempts} attempts. Aborting the installation.")

    def prompt_for_volume_setup(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.READ_VOLUME, Privilege.WRITE_VOLUME},
        ),
        max_attempts: int = 3,
    ) -> str:
        for _ in range(1, max_attempts + 1):
            volume_name = self._prompts.question("Enter volume name", default=default_volume_name)
            volume = self._catalog_ops.get_volume(catalog, schema, volume_name)
            if volume:
                logger.info(f"Found existing volume `{volume_name}` in catalog `{catalog}` and schema `{schema}`")
                user_name = self._user.user_name
                assert user_name is not None
                if self._catalog_ops.has_volume_access(volume, user_name, required_privileges):
                    return volume_name
                logger.info(
                    f"User `{user_name}` doesn't have privilege to use volume `{volume_name}` "
                    f"in catalog `{catalog}` and schema `{schema}`"
                )
                if not self._prompts.confirm("Do you want to use another volume?"):
                    raise SystemExit("Aborting the installation.")
            else:
                if self._prompts.confirm(
                    f"Volume `{volume_name}` doesn't exist in catalog `{catalog}` and schema `{schema}`. Create it?"
                ):
                    result = self._catalog_ops.create_volume(catalog, schema, volume_name)
                    assert result.name is not None
                    return result.name
                raise SystemExit("Aborting the installation.")
        raise SystemExit(f"Couldn't get answer within {max_attempts} attempts. Aborting the installation.")

    def prompt_for_warehouse_setup(self, warehouse_name_prefix: str) -> str:
        def warehouse_type(_):
            return _.warehouse_type.value if not _.enable_serverless_compute else "SERVERLESS"

        pro_warehouses = {"[Create new PRO SQL warehouse]": "create_new"} | {
            f"{_.name} ({_.id}, {warehouse_type(_)}, {_.state.value})": _.id
            for _ in self._ws.warehouses.list()
            if _.warehouse_type == EndpointInfoWarehouseType.PRO
        }
        warehouse_id = self._prompts.choice_from_dict(
            "Select PRO or SERVERLESS SQL warehouse",
            pro_warehouses,
        )
        if warehouse_id == "create_new":
            new_warehouse = self._ws.warehouses.create(
                name=f"{warehouse_name_prefix} {time.time_ns()}",
                spot_instance_policy=SpotInstancePolicy.COST_OPTIMIZED,
                warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
                cluster_size="Small",
                max_num_clusters=1,
            )
            warehouse_id = new_warehouse.id
        return warehouse_id



from pyspark.sql import SparkSession
from sqlglot import Dialect

from databricks.labs.remorph.reconcile.connectors.data_source import DataSource
from databricks.labs.remorph.reconcile.connectors.databricks import DatabricksDataSource
from databricks.labs.remorph.reconcile.connectors.oracle import OracleDataSource
from databricks.labs.remorph.reconcile.connectors.snowflake import SnowflakeDataSource
from databricks.labs.remorph.snow.databricks import Databricks
from databricks.labs.remorph.snow.oracle import Oracle
from databricks.labs.remorph.snow.snowflake import Snow
from databricks.sdk import WorkspaceClient


def create_adapter(
    engine: Dialect,
    spark: SparkSession,
    ws: WorkspaceClient,
    secret_scope: str,
) -> DataSource:
    if isinstance(engine, Snow):
        return SnowflakeDataSource(engine, spark, ws, secret_scope)
    if isinstance(engine, Oracle):
        return OracleDataSource(engine, spark, ws, secret_scope)
    if isinstance(engine, Databricks):
        return DatabricksDataSource(engine, spark, ws, secret_scope)
    raise ValueError(f"Unsupported source type --> {engine}")



import dataclasses
import logging
import os
import webbrowser

from databricks.labs.blueprint.entrypoint import get_logger, is_in_debug
from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installation import SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, PermissionDenied
from databricks.sdk.service.catalog import Privilege

from databricks.labs.remorph.__about__ import __version__
from databricks.labs.remorph.config import (
    MorphConfig,
    ReconcileConfig,
    SQLGLOT_DIALECTS,
    DatabaseConfig,
    RemorphConfigs,
    ReconcileMetadataConfig,
)
from databricks.labs.remorph.contexts.application import ApplicationContext
from databricks.labs.remorph.deployment.configurator import ResourceConfigurator
from databricks.labs.remorph.deployment.installation import WorkspaceInstallation
from databricks.labs.remorph.reconcile.constants import ReconReportType, ReconSourceType

logger = logging.getLogger(__name__)

TRANSPILER_WAREHOUSE_PREFIX = "Remorph Transpiler Validation"
MODULES = sorted({"transpile", "reconcile", "all"})


class WorkspaceInstaller:
    def __init__(
        self,
        ws: WorkspaceClient,
        prompts: Prompts,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        resource_configurator: ResourceConfigurator,
        workspace_installation: WorkspaceInstallation,
        environ: dict[str, str] | None = None,
    ):
        self._ws = ws
        self._prompts = prompts
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._resource_configurator = resource_configurator
        self._ws_installation = workspace_installation

        if not environ:
            environ = dict(os.environ.items())

        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

    def run(
        self,
        config: RemorphConfigs | None = None,
    ) -> RemorphConfigs:
        logger.info(f"Installing Remorph v{self._product_info.version()}")
        if not config:
            config = self.configure()
        if self._is_testing():
            return config
        self._ws_installation.install(config)
        logger.info("Installation completed successfully! Please refer to the documentation for the next steps.")
        return config

    def configure(self, module: str | None = None) -> RemorphConfigs:
        selected_module = module or self._prompts.choice("Select a module to configure:", MODULES)
        match selected_module:
            case "transpile":
                logger.info("Configuring remorph `transpile`.")
                return RemorphConfigs(self._configure_transpile(), None)
            case "reconcile":
                logger.info("Configuring remorph `reconcile`.")
                return RemorphConfigs(None, self._configure_reconcile())
            case "all":
                logger.info("Configuring remorph `transpile` and `reconcile`.")
                return RemorphConfigs(
                    self._configure_transpile(),
                    self._configure_reconcile(),
                )
            case _:
                raise ValueError(f"Invalid input: {selected_module}")

    def _is_testing(self):
        return self._product_info.product_name() != "remorph"

    def _configure_transpile(self) -> MorphConfig:
        try:
            self._installation.load(MorphConfig)
            logger.info("Remorph `transpile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                raise SystemExit(
                    "Remorph `transpile` is already installed and no override has been requested. Exiting..."
                )
        except NotFound:
            logger.info("Couldn't find existing `transpile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `transpile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_transpile_installation()
        logger.info("Finished configuring remorph `transpile`.")
        return config

    def _configure_new_transpile_installation(self) -> MorphConfig:
        default_config = self._prompt_for_new_transpile_installation()
        runtime_config = None
        catalog_name = "remorph"
        schema_name = "transpiler"
        if not default_config.skip_validation:
            catalog_name = self._configure_catalog()
            schema_name = self._configure_schema(catalog_name, "transpile")
            runtime_config = self._configure_runtime()

        config = dataclasses.replace(
            default_config,
            catalog_name=catalog_name,
            schema_name=schema_name,
            sdk_config=runtime_config,
        )
        self._save_config(config)
        return config

    def _prompt_for_new_transpile_installation(self) -> MorphConfig:
        logger.info("Please answer a few questions to configure remorph `transpile`")
        source = self._prompts.choice("Select the source:", list(SQLGLOT_DIALECTS.keys()))
        input_sql = self._prompts.question("Enter input SQL path (directory/file)")
        output_folder = self._prompts.question("Enter output directory", default="transpiled")
        run_validation = self._prompts.confirm(
            "Would you like to validate the syntax and semantics of the transpiled queries?"
        )

        return MorphConfig(
            source=source,
            skip_validation=(not run_validation),
            mode="current",  # mode will not have a prompt as this is a hidden flag
            input_sql=input_sql,
            output_folder=output_folder,
        )

    def _configure_catalog(
        self,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_CATALOG},
        ),
        max_attempts: int = 3,
    ) -> str:
        return self._resource_configurator.prompt_for_catalog_setup(required_privileges, max_attempts)

    def _configure_schema(
        self,
        catalog: str,
        default_schema_name: str,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.USE_SCHEMA},
        ),
        max_attempts: int = 3,
    ) -> str:
        return self._resource_configurator.prompt_for_schema_setup(
            catalog,
            default_schema_name,
            required_privileges,
            max_attempts,
        )

    def _configure_runtime(self) -> dict[str, str]:
        if self._prompts.confirm("Do you want to use SQL Warehouse for validation?"):
            warehouse_id = self._resource_configurator.prompt_for_warehouse_setup(TRANSPILER_WAREHOUSE_PREFIX)
            return {"warehouse_id": warehouse_id}

        if self._ws.config.cluster_id:
            logger.info(f"Using cluster {self._ws.config.cluster_id} for validation")
            return {"cluster_id": self._ws.config.cluster_id}

        cluster_id = self._prompts.question("Enter a valid cluster_id to proceed")
        return {"cluster_id": cluster_id}

    def _configure_reconcile(self) -> ReconcileConfig:
        try:
            self._installation.load(ReconcileConfig)
            logger.info("Remorph `reconcile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                raise SystemExit(
                    "Remorph `reconcile` is already installed and no override has been requested. Exiting..."
                )
        except NotFound:
            logger.info("Couldn't find existing `reconcile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `reconcile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_reconcile_installation()
        logger.info("Finished configuring remorph `reconcile`.")
        return config

    def _configure_new_reconcile_installation(self) -> ReconcileConfig:
        default_config = self._prompt_for_new_reconcile_installation()
        self._save_config(default_config)
        return default_config

    def _prompt_for_new_reconcile_installation(self) -> ReconcileConfig:
        logger.info("Please answer a few questions to configure remorph `reconcile`")
        data_source = self._prompts.choice(
            "Select the Data Source:", [source_type.value for source_type in ReconSourceType]
        )
        report_type = self._prompts.choice(
            "Select the report type:", [report_type.value for report_type in ReconReportType]
        )
        scope_name = self._prompts.question(
            f"Enter Secret scope name to store `{data_source.capitalize()}` connection details / secrets",
            default=f"remorph_{data_source}",
        )

        db_config = self._prompt_for_reconcile_database_config(data_source)
        metadata_config = self._prompt_for_reconcile_metadata_config()

        return ReconcileConfig(
            data_source=data_source,
            report_type=report_type,
            secret_scope=scope_name,
            database_config=db_config,
            metadata_config=metadata_config,
        )

    def _prompt_for_reconcile_database_config(self, source) -> DatabaseConfig:
        source_catalog = None
        if source == ReconSourceType.SNOWFLAKE.value:
            source_catalog = self._prompts.question(f"Enter source catalog name for `{source.capitalize()}`")

        schema_prompt = f"Enter source schema name for `{source.capitalize()}`"
        if source in {ReconSourceType.ORACLE.value}:
            schema_prompt = f"Enter source database name for `{source.capitalize()}`"

        source_schema = self._prompts.question(schema_prompt)
        target_catalog = self._prompts.question("Enter target catalog name for Databricks")
        target_schema = self._prompts.question("Enter target schema name for Databricks")

        return DatabaseConfig(
            source_schema=source_schema,
            target_catalog=target_catalog,
            target_schema=target_schema,
            source_catalog=source_catalog,
        )

    def _prompt_for_reconcile_metadata_config(self) -> ReconcileMetadataConfig:
        logger.info("Configuring reconcile metadata.")
        catalog = self._configure_catalog()
        schema = self._configure_schema(
            catalog,
            "reconcile",
            required_privileges=(
                {Privilege.ALL_PRIVILEGES},
                {Privilege.USE_SCHEMA, Privilege.CREATE_VOLUME},
            ),
        )
        volume = self._configure_volume(catalog, schema, "reconcile_volume")
        return ReconcileMetadataConfig(catalog=catalog, schema=schema, volume=volume)

    def _configure_volume(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
        required_privileges: tuple[set[Privilege], ...] = (
            {Privilege.ALL_PRIVILEGES},
            {Privilege.READ_VOLUME, Privilege.WRITE_VOLUME},
        ),
        max_attempts: int = 3,
    ) -> str:
        return self._resource_configurator.prompt_for_volume_setup(
            catalog,
            schema,
            default_volume_name,
            required_privileges,
            max_attempts,
        )

    def _save_config(self, config: MorphConfig | ReconcileConfig):
        logger.info(f"Saving configuration file {config.__file__}")
        self._installation.save(config)
        ws_file_url = self._installation.workspace_link(config.__file__)
        if self._prompts.confirm(f"Open config file {ws_file_url} in the browser?"):
            webbrowser.open(ws_file_url)


if __name__ == "__main__":
    logger = get_logger(__file__)
    logger.setLevel("INFO")
    if is_in_debug():
        logging.getLogger("databricks").setLevel(logging.DEBUG)

    app_context = ApplicationContext(WorkspaceClient(product="remorph", product_version=__version__))
    installer = WorkspaceInstaller(
        app_context.workspace_client,
        app_context.prompts,
        app_context.installation,
        app_context.install_state,
        app_context.product_info,
        app_context.resource_configurator,
        app_context.workspace_installation,
    )
    installer.run()



import json
import logging
from datetime import timedelta
from importlib.abc import Traversable
from typing import Any

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError
from databricks.sdk.errors import InvalidParameterValue
from databricks.sdk.retries import retried
from databricks.sdk.service.dashboards import Dashboard

logger = logging.getLogger(__name__)


class DashboardDeployment:
    _UPLOAD_TIMEOUT = timedelta(seconds=30)

    def __init__(self, ws: WorkspaceClient, installation: Installation, install_state: InstallState):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state

    def deploy(self, name: str, dashboard_file: Traversable, parameters: dict[str, Any] | None = None):
        logger.debug(f"Deploying dashboard {name} from {dashboard_file.name}")
        dashboard_data = self._substitute_params(dashboard_file, parameters or {})
        dashboard = self._update_or_create_dashboard(name, dashboard_data, dashboard_file)
        logger.info(f"Dashboard deployed with dashboard_id {dashboard.dashboard_id}")
        logger.info(f"Dashboard URL: {self._ws.config.host}/sql/dashboardsv3/{dashboard.dashboard_id}")
        self._install_state.save()

    @retried(on=[DatabricksError], timeout=_UPLOAD_TIMEOUT)
    def _update_or_create_dashboard(self, name: str, dashboard_data, dashboard_file) -> Dashboard:
        if name in self._install_state.dashboards:
            try:
                dashboard_id = self._install_state.dashboards[name]
                logger.info(f"Updating dashboard with id={dashboard_id}")
                updated_dashboard = self._ws.lakeview.update(
                    dashboard_id,
                    display_name=self._name_with_prefix(name),
                    serialized_dashboard=dashboard_data,
                )
                return updated_dashboard
            except InvalidParameterValue:
                del self._install_state.dashboards[name]
                logger.warning(f"Dashboard {name} does not exist anymore for some reason.")
                return self._update_or_create_dashboard(name, dashboard_data, dashboard_file)
        logger.info(f"Creating new dashboard {name}")
        new_dashboard = self._ws.lakeview.create(
            display_name=self._name_with_prefix(name),
            parent_path=self._install_state.install_folder(),
            serialized_dashboard=dashboard_data,
        )
        assert new_dashboard.dashboard_id is not None
        self._install_state.dashboards[name] = new_dashboard.dashboard_id
        return new_dashboard

    def _substitute_params(self, dashboard_file: Traversable, parameters: dict[str, Any]) -> str:
        if not parameters:
            return dashboard_file.read_text()

        with dashboard_file.open() as f:
            dashboard_data = json.load(f)

        for dataset in dashboard_data.get("datasets", []):
            for param in dataset.get("parameters", []):
                if param["keyword"] in parameters:
                    param["defaultSelection"] = {
                        "values": {
                            "dataType": "STRING",
                            "values": [
                                {"value": parameters[param["keyword"]]},
                            ],
                        },
                    }

        return json.dumps(dashboard_data)

    def _name_with_prefix(self, name: str) -> str:
        prefix = self._installation.product()
        return f"[{prefix.upper()}] {name}"



import logging
from dataclasses import dataclass

from sqlglot.dialects.dialect import Dialect, Dialects, DialectType

from databricks.labs.remorph.helpers.morph_status import ParserError
from databricks.labs.remorph.reconcile.recon_config import Table
from databricks.labs.remorph.snow import databricks, experimental, oracle, snowflake, presto

logger = logging.getLogger(__name__)

SQLGLOT_DIALECTS: dict[str, DialectType] = {
    "athena": Dialects.ATHENA,
    "bigquery": Dialects.BIGQUERY,
    "databricks": databricks.Databricks,
    "experimental": experimental.DatabricksExperimental,
    "mysql": Dialects.MYSQL,
    "netezza": Dialects.POSTGRES,
    "oracle": oracle.Oracle,
    "postgresql": Dialects.POSTGRES,
    "presto": presto.Presto,
    "redshift": Dialects.REDSHIFT,
    "snowflake": snowflake.Snow,
    "sqlite": Dialects.SQLITE,
    "teradata": Dialects.TERADATA,
    "trino": Dialects.TRINO,
    "tsql": Dialects.TSQL,
    "vertica": Dialects.POSTGRES,
}


def get_dialect(engine: str) -> Dialect:
    return Dialect.get_or_raise(SQLGLOT_DIALECTS.get(engine))


def get_key_from_dialect(input_dialect: Dialect) -> str:
    return [source_key for source_key, dialect in SQLGLOT_DIALECTS.items() if dialect == input_dialect][0]


@dataclass
class MorphConfig:
    __file__ = "config.yml"
    __version__ = 1

    source: str
    sdk_config: dict[str, str] | None = None
    input_sql: str | None = None
    output_folder: str | None = None
    skip_validation: bool = False
    catalog_name: str = "remorph"
    schema_name: str = "transpiler"
    mode: str = "current"

    def get_read_dialect(self):
        return get_dialect(self.source)

    def get_write_dialect(self):
        if self.mode == "experimental":
            return get_dialect("experimental")
        return get_dialect("databricks")


@dataclass
class TableRecon:
    __file__ = "recon_config.yml"
    __version__ = 1

    source_schema: str
    target_catalog: str
    target_schema: str
    tables: list[Table]
    source_catalog: str | None = None

    def __post_init__(self):
        self.source_schema = self.source_schema.lower()
        self.target_schema = self.target_schema.lower()
        self.target_catalog = self.target_catalog.lower()
        self.source_catalog = self.source_catalog.lower() if self.source_catalog else self.source_catalog


@dataclass
class DatabaseConfig:
    source_schema: str
    target_catalog: str
    target_schema: str
    source_catalog: str | None = None


@dataclass
class TranspilationResult:
    transpiled_sql: list[str]
    parse_error_list: list[ParserError]


@dataclass
class ValidationResult:
    validated_sql: str
    exception_msg: str | None


@dataclass
class ReconcileTablesConfig:
    filter_type: str  # all/include/exclude
    tables_list: list[str]  # [*, table1, table2]


@dataclass
class ReconcileMetadataConfig:
    catalog: str = "remorph"
    schema: str = "reconcile"
    volume: str = "reconcile_volume"


@dataclass
class ReconcileConfig:
    __file__ = "reconcile.yml"
    __version__ = 1

    data_source: str
    report_type: str
    secret_scope: str
    database_config: DatabaseConfig
    metadata_config: ReconcileMetadataConfig
    job_id: str | None = None
    tables: ReconcileTablesConfig | None = None


@dataclass
class RemorphConfigs:
    morph: MorphConfig | None = None
    reconcile: ReconcileConfig | None = None
